"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BindingBuilder = exports.SPECIAL_REFERENCES = exports.PARENTS_REFERENCE_NAME = exports.PARENT_REFERENCE_NAME = exports.VALUE_REFERENCE_NAME = exports.VIEW_MODEL_REFERENCE_NAME = exports.BINDING_CONTEXT_PROPERTY_NAME = void 0;
const parser = __importStar(require("@babel/parser"));
const traverse_1 = __importDefault(require("@babel/traverse"));
const t = __importStar(require("@babel/types"));
const helpers_1 = require("../helpers");
const BINDING_REGEX = /\{{(.+?)\}}/;
const CONVERTER_EXPRESSION_SEPARATOR = '|';
const BINDING_AST_VALIDATORS = [
    t.isArrayExpression,
    t.isBinaryExpression,
    t.isCallExpression,
    t.isConditionalExpression,
    t.isIdentifier,
    t.isLiteral,
    t.isLogicalExpression,
    t.isMemberExpression,
    t.isNewExpression,
    t.isObjectExpression,
    t.isOptionalCallExpression,
    t.isOptionalMemberExpression,
    t.isProperty,
    t.isSpreadElement,
    t.isTemplateElement,
    t.isUnaryExpression
];
exports.BINDING_CONTEXT_PROPERTY_NAME = 'bindingContext';
exports.VIEW_MODEL_REFERENCE_NAME = 'viewModel';
exports.VALUE_REFERENCE_NAME = '$value';
exports.PARENT_REFERENCE_NAME = '$parent';
exports.PARENTS_REFERENCE_NAME = '$parents';
exports.SPECIAL_REFERENCES = [
    exports.VALUE_REFERENCE_NAME,
    exports.PARENT_REFERENCE_NAME,
    exports.PARENTS_REFERENCE_NAME
];
class BindingBuilder {
    isBindingValue(value) {
        return BINDING_REGEX.test(value);
    }
    convertValueToBindingOptions(propertyDetails) {
        const code = this.getBindingCode(propertyDetails.value);
        if (!code?.length) {
            throw new Error('Invalid binding expression. Curly brackets are empty');
        }
        const ast = parser.parse(code);
        if (ast.program.body.length !== 1 || !t.isExpressionStatement(ast.program.body[0])) {
            throw new Error(`Invalid binding expression. Binding must be a single-line expression statement: ${propertyDetails.value}`);
        }
        const expressionStatement = ast.program.body[0];
        const bindingOptions = {
            viewPropertyDetails: propertyDetails,
            properties: [],
            astExpression: null,
            isTwoWay: !propertyDetails.isEventListener && !propertyDetails.isSubProperty && (this.isTwoWayBindingExpression(expressionStatement.expression)
                || this.isConverterExpression(expressionStatement.expression)
                    && this.isTwoWayBindingExpression(expressionStatement.expression.left)),
            specialReferenceCount: 0,
            parentKeyAstExpressions: [],
            get [Symbol.toStringTag]() {
                return propertyDetails.value;
            }
        };
        (0, traverse_1.default)(expressionStatement, {
            noScope: true,
            // Validate AST
            enter: (path) => {
                this.checkIfBindingExpressionIsValid(path, bindingOptions);
            },
            // We apply optional chaining by default in order to get rid of binding errors when binding context is not set early enough
            CallExpression: (path) => {
                const node = path.node;
                path.replaceWith(t.optionalCallExpression(node.callee, node.arguments, true));
            },
            // Traverse through all identifiers and keep track of the ones that should be observed
            Identifier: (path) => {
                this.handleIdentifier(path, bindingOptions);
            },
            // We apply optional chaining by default in order to get rid of binding errors when binding context is not set early enough
            MemberExpression: (path) => {
                const node = path.node;
                path.replaceWith(t.optionalMemberExpression(node.object, node.property, node.computed, true));
            },
            exit: (path) => {
                const parentPath = path.parentPath;
                if (parentPath != null) {
                    // In the case of converters, we ensure expression is a two-way binding by checking left-side reference
                    if (bindingOptions.isTwoWay && !bindingOptions.properties.length && this.isConverterExpression(parentPath.node) && path.node === parentPath.node.left) {
                        bindingOptions.isTwoWay = false;
                    }
                }
                /**
                 * Always handle converters after they get fully traversed.
                 * It helps with the order of bindable properties and builder does not append 'viewModel' caller to them.
                 */
                if (this.isConverterExpression(path.node)) {
                    this.handleConverter(path, bindingOptions);
                }
            }
        });
        // If property is an event listener, make sure that listener function will use view model as context
        if (bindingOptions.viewPropertyDetails.isEventListener) {
            bindingOptions.astExpression = this.getEventExpressionWithContext(expressionStatement.expression);
        }
        else {
            bindingOptions.astExpression = expressionStatement.expression;
        }
        // Ensure there are properties as first one is always needed for setting two-way binding value
        if (!bindingOptions.properties.length && bindingOptions.isTwoWay) {
            bindingOptions.isTwoWay = false;
            delete bindingOptions.converterToModelAstExpression;
        }
        return bindingOptions;
    }
    getEventExpressionWithContext(astExpression) {
        return t.optionalCallExpression(t.optionalMemberExpression(astExpression, t.identifier('bind'), false, true), [
            t.identifier(exports.VIEW_MODEL_REFERENCE_NAME)
        ], true);
    }
    checkIfBindingExpressionIsValid(ast, bindingOptions) {
        for (const validate of BINDING_AST_VALIDATORS) {
            if (validate(ast)) {
                return;
            }
        }
        throw new Error(`Invalid binding expression: ${bindingOptions.toString()}`);
    }
    handleIdentifier(path, bindingOptions) {
        const node = path.node;
        let parentNode;
        let isBindingProperty = false;
        if (path.parentPath != null) {
            parentNode = path.parentPath.node;
            if (this.isMemberExpression(parentNode)) {
                // Identifiers that are first properties or properties inside a computed property
                if (node === parentNode.object || parentNode.computed) {
                    isBindingProperty = true;
                }
            }
            else if (t.isObjectProperty(parentNode)) {
                // Identifiers that are used as object values or properties inside a computed object key
                if (node === parentNode.value || parentNode.computed) {
                    isBindingProperty = true;
                }
            }
            else {
                isBindingProperty = true;
            }
        }
        else {
            isBindingProperty = true;
        }
        if (isBindingProperty) {
            // This handles identifiers like $value, $parent, $parents, etc
            if (exports.SPECIAL_REFERENCES.includes(node.name)) {
                if (node.name === exports.PARENTS_REFERENCE_NAME) {
                    const parentKeyAst = this.getAstForParentKey(parentNode, bindingOptions.toString());
                    bindingOptions.parentKeyAstExpressions.push(parentKeyAst);
                }
                bindingOptions.specialReferenceCount++;
            }
            else {
                // Avoid keeping track of the same identifier twice
                if (!bindingOptions.properties.includes(path.node.name)) {
                    bindingOptions.properties.push(path.node.name);
                }
                // Append a view model reference for later use
                path.replaceWith(t.memberExpression(t.identifier(exports.VIEW_MODEL_REFERENCE_NAME), path.node));
                path.skip();
            }
        }
    }
    handleConverter(path, bindingOptions) {
        const node = path.node;
        const isRightSideCallExpression = this.isCallExpression(node.right);
        const isRightSideExpressionValid = t.isIdentifier(node.right) || this.isMemberExpression(node.right) || isRightSideCallExpression;
        // Right-side has to be a property or function call
        if (!isRightSideExpressionValid) {
            throw new Error(`Invalid right-side reference for converter. Expression: ${bindingOptions.toString()}`);
        }
        // Keep callee and arguments from converter expression
        let callee;
        const args = [];
        if (isRightSideCallExpression) {
            const callExpression = node.right;
            callee = callExpression.callee;
            if (callExpression.arguments.length) {
                args.push(...callExpression.arguments);
            }
            else {
                // For HMR to be able to tell the difference between myConverter() and myConverter, we append an undefined parameter in the case of call expression
                args.push(t.identifier('undefined'));
            }
        }
        else {
            callee = node.right;
        }
        // This is an expression for using converter to set value from view to view model in case converter is the binding expression itself
        if (path.parentPath == null && bindingOptions.isTwoWay) {
            bindingOptions.converterToModelAstExpression = t.sequenceExpression([
                node.left,
                t.callExpression(t.memberExpression(t.memberExpression(t.identifier('global'), t.identifier(helpers_1.GLOBAL_UI_REF)), t.identifier('runConverterCallback')), [
                    t.identifier(exports.VIEW_MODEL_REFERENCE_NAME),
                    callee,
                    t.arrayExpression([
                        t.memberExpression(t.identifier('args'), t.identifier('value')),
                        ...args
                    ]),
                    t.booleanLiteral(true)
                ])
            ]);
        }
        // Generate a function call that executes converter in runtime
        path.replaceWith(t.callExpression(t.memberExpression(t.memberExpression(t.identifier('global'), t.identifier(helpers_1.GLOBAL_UI_REF)), t.identifier('runConverterCallback')), [
            t.identifier(exports.VIEW_MODEL_REFERENCE_NAME),
            callee,
            t.arrayExpression([
                node.left,
                ...args
            ])
        ]));
        path.skip();
    }
    getAstForParentKey(parentNode, expressionDesc) {
        let propertyExpression;
        if (parentNode == null || !this.isMemberExpression(parentNode)) {
            throw new Error(`Invalid '${exports.PARENTS_REFERENCE_NAME}' expression reference. No element name has been given to search for: ${expressionDesc}`);
        }
        if (parentNode.computed) {
            propertyExpression = parentNode.property;
        }
        else {
            if (t.isIdentifier(parentNode.property)) {
                propertyExpression = t.stringLiteral(parentNode.property.name);
            }
            else {
                throw new Error(`Invalid '${exports.PARENTS_REFERENCE_NAME}' expression reference. Invalid element name: ${expressionDesc}`);
            }
        }
        return propertyExpression;
    }
    getBindingCode(bindingValue) {
        const matchResult = bindingValue.match(BINDING_REGEX);
        if (matchResult == null) {
            throw new Error(`Cannot retrieve code content from non-binding value: ${bindingValue}`);
        }
        return matchResult[1].trim();
    }
    isCallExpression(ast) {
        return t.isCallExpression(ast) || t.isOptionalCallExpression(ast);
    }
    isConverterExpression(ast) {
        return t.isBinaryExpression(ast, {
            operator: CONVERTER_EXPRESSION_SEPARATOR
        });
    }
    isMemberExpression(ast) {
        return t.isMemberExpression(ast) || t.isOptionalMemberExpression(ast);
    }
    isTwoWayBindingExpression(ast) {
        return t.isIdentifier(ast) || this.isMemberExpression(ast);
    }
}
exports.BindingBuilder = BindingBuilder;
//# sourceMappingURL=binding-builder.js.map